/* gmpc-discogs (GMPC plugin)
 * Copyright (C) 2008-2009 Qball Cow <qball@sarine.nl>
 * Project homepage: http://gmpcwiki.sarine.nl/
 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
/* gtk and lib stuff */
#include <gtk/gtk.h>
#include <glib.h>
#include <glib/gstdio.h>

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libmpd/debug_printf.h>

#include <gmpc/plugin.h>
#include <gmpc/gmpc_easy_download.h>
#include <gmpc/metadata.h>

#include <config.h>

#define DISCOGS_API_KEY "332020810c"
#define DISCOGS_API_ROOT "http://www.discogs.com/"

static int discogs_get_enabled()
{
	return cfg_get_single_value_as_int_with_default(config, "cover-discogs", "enable", TRUE);
}
static void discogs_set_enabled(int enabled)
{
	cfg_set_single_value_as_int(config, "cover-discogs", "enable", enabled);
}
static int discogs_fetch_cover_priority(void){
	return cfg_get_single_value_as_int_with_default(config, "cover-discogs", "priority", 80);
}
static void discogs_fetch_cover_priority_set(int priority){
	cfg_set_single_value_as_int(config, "cover-discogs", "priority", priority);
}

static xmlNodePtr get_first_node_by_name(xmlNodePtr xml, gchar *name) {
    if(name == NULL) return NULL;
	if(xml) {
		xmlNodePtr c = xml->xmlChildrenNode;
		for(;c;c=c->next) {
			if(c->name && xmlStrEqual(c->name, (xmlChar *) name))
				return c;
		}
	}
	return NULL;
}
static int __discogs_art_get_artist_image(mpd_Song *song,char *nartist, char **path)
{
    gchar *secondary = NULL;
    gchar *primary = NULL;
    glong psize = 0;
    glong ssize = 0;

    gchar *artist_uri = NULL;
    gmpc_easy_download_struct data= {NULL, 0,-1,NULL, NULL};
	int found = META_DATA_UNAVAILABLE;
	char furl[1024];

	snprintf(furl,1024,DISCOGS_API_ROOT"search?type=all&f=xml&q=%s&api_key=%s", nartist,DISCOGS_API_KEY);
    if(gmpc_easy_download(furl, &data))
    {
        xmlDocPtr doc;
        if(data.size < 4 || strncmp(data.data, "<res",4))
        {
            printf("1: Invalid xml: %s\n", furl);
            gmpc_easy_download_clean(&data);
        }
        else
        {
            /**
             * Get artist name
             */
            doc = xmlParseMemory(data.data,data.size);
            gmpc_easy_download_clean(&data);
            if(doc)
            {
                xmlNodePtr root = xmlDocGetRootElement(doc);
                if(root)
                {
                    /* loop through all albums */
                    xmlNodePtr cur = get_first_node_by_name(root,"exactresults");
                    if(cur)
                    {
                        xmlNodePtr cur2 = get_first_node_by_name(cur,"result");
                        if(cur2) {
                            xmlNodePtr cur3 = get_first_node_by_name(cur2,"uri");
                            if(cur3){
                                xmlChar *xurl = xmlNodeGetContent(cur3);
                                artist_uri = g_strdup((char *)xurl);
                                xmlFree(xurl);
                            }
                        }
                    }
                }
                xmlFreeDoc(doc);
            }
        }
    }
    if(artist_uri) {
        if(strstr(artist_uri, "?") != NULL)
        {
            snprintf(furl,1024,"%s&f=xml&api_key=%s", artist_uri,DISCOGS_API_KEY);
        }else{
            snprintf(furl,1024,"%s?f=xml&api_key=%s", artist_uri,DISCOGS_API_KEY);
        }
        if(gmpc_easy_download(furl, &data))
        {
            xmlDocPtr doc;
            if(data.size < 4 || strncmp(data.data, "<res",4))
            {
                printf("2: Invalid xml: %s\n", furl);
                gmpc_easy_download_clean(&data);
            }
            else
            {
                doc = xmlParseMemory(data.data,data.size);
                gmpc_easy_download_clean(&data);
                if(doc)
                {
                    xmlNodePtr root = xmlDocGetRootElement(doc);
                    if(root)
                    {
                        /* loop through all albums */
                        xmlNodePtr cur = get_first_node_by_name(root,"artist");
                        if(cur)
                        {
                            xmlNodePtr cur2 = get_first_node_by_name(cur,"images");
                            if(cur2) {
                                xmlNodePtr cur3 = get_first_node_by_name(cur2,"image");
                                while(cur3 && *path == NULL){
                                    xmlChar *temp = xmlGetProp(cur3, (xmlChar *)"type");
                                    if(temp){
                                        if(xmlStrEqual(temp, (xmlChar *)"primary"))
                                        {
                                            xmlChar *xurl = xmlGetProp(cur3, (xmlChar *)"uri");
                                            xmlChar *size = xmlGetProp(cur3, (xmlChar *)"height");
                                            glong s = (gulong)g_ascii_strtoll((gchar *)size,NULL, 10);
                                            if(s > psize) {
                                                psize = s;
                                                if(primary) g_free(secondary);
                                                primary = g_strdup((gchar *)xurl);
                                            }
                                            if(xurl) xmlFree(xurl);
                                            if(size) xmlFree(size);
                                        } else if(xmlStrEqual(temp, (xmlChar *)"secondary"))
                                        {
                                            xmlChar *xurl = xmlGetProp(cur3, (xmlChar *)"uri");
                                            xmlChar *size = xmlGetProp(cur3, (xmlChar *)"height");
                                            glong s = (gulong)g_ascii_strtoll((gchar *)size,NULL, 10);
                                            if(s > ssize) {
                                                ssize = s;
                                                if(secondary) g_free(secondary);
                                                secondary = g_strdup((gchar *)xurl);
                                            }
                                            if(xurl) xmlFree(xurl);
                                            if(size) xmlFree(size);
                                        }

                                        xmlFree(temp);
                                    }
                                    cur3 = cur3->next;
                                }
                            }
                        }
                    }
                    xmlFreeDoc(doc);
                }
            }
            gmpc_easy_download_clean(&data);
        }
    }
    if(artist_uri) g_free(artist_uri);
    if(primary || secondary)
    {
        gmpc_easy_download((const char *)(primary)?primary:secondary, &data);
        if(data.size){
            FILE *fp = NULL;
            char *imgpath = NULL;
            imgpath = gmpc_get_metadata_filename(META_ARTIST_ART, song, "jpg");
            fp = g_fopen(imgpath, "wb");
            if(fp)
            {
                size_t size;
                size = fwrite(data.data, sizeof(char), data.size,fp);
                if(size != data.size){
                    debug_printf(DEBUG_ERROR, "Something went wrong writing image to file");
                }
                fclose(fp);
            }
            *path = imgpath;
            found = META_DATA_AVAILABLE;
            debug_printf(DEBUG_INFO,"Found cover art, returning\n");
        }
        gmpc_easy_download_clean(&data);
    }
    if(primary) g_free(primary);
    if(secondary) g_free(secondary);
	return found;
}

static int __discogs_art_get_album_image(mpd_Song *song,char *nartist,char *nalbum, char **path)
{
    gchar *secondary = NULL;
    gchar *primary = NULL;
    glong psize = 0;
    glong ssize = 0;


    gchar *artist_uri = NULL;
    gmpc_easy_download_struct data= {NULL, 0,-1,NULL, NULL};
	int found = META_DATA_UNAVAILABLE;
	char furl[1024];


	snprintf(furl,1024,DISCOGS_API_ROOT"search?type=all&f=xml&q=%s%%20%s&api_key=%s", nartist,nalbum,DISCOGS_API_KEY);
    if(gmpc_easy_download(furl, &data))
    {
        xmlDocPtr doc;
        /**
         * Get artist name
         */
        if(data.size < 4 || strncmp(data.data, "<res",4))
        {
            printf("3: Invalid xml: %s\n", furl);
            gmpc_easy_download_clean(&data);
        }
        else
        {
            doc = xmlParseMemory(data.data,data.size);
            gmpc_easy_download_clean(&data);
            if(doc)
            {
                xmlNodePtr root = xmlDocGetRootElement(doc);
                if(root)
                {
                    /* loop through all albums */
                    xmlNodePtr cur = get_first_node_by_name(root,"searchresults");
                    if(cur)
                    {
                        xmlNodePtr cur2 = get_first_node_by_name(cur,"result");
                        if(cur2) {
                            xmlNodePtr cur4 = get_first_node_by_name(cur2,"title");
                            if(cur4){
                                xmlChar *title = xmlNodeGetContent(cur4);
                                /** Todo make this check fuzzy */
                                if(title && strstr((char *)title, song->album))
                                {
                                    xmlNodePtr cur3 = get_first_node_by_name(cur2,"uri");
                                    if(cur3){
                                        xmlChar *xurl = xmlNodeGetContent(cur3);
                                        artist_uri = g_strdup((char *)xurl);
                                        xmlFree(xurl);
                                    }
                                }
                                if(title)xmlFree(title);
                            }
                        }
                    }
                }
                xmlFreeDoc(doc);
            }
        }
    }
    if(artist_uri) {
        snprintf(furl,1024,"%s?f=xml&api_key=%s", artist_uri,DISCOGS_API_KEY);

        if(gmpc_easy_download(furl, &data))
        {
            xmlDocPtr doc;
            if(data.size < 4 || strncmp(data.data, "<res",4))
            {
                printf("4: Invalid xml: %s\n", furl);
                gmpc_easy_download_clean(&data);
            }
            else
            {

                doc = xmlParseMemory(data.data,data.size);
                gmpc_easy_download_clean(&data);
                if(doc)
                {
                    xmlNodePtr root = xmlDocGetRootElement(doc);
                    if(root)
                    {
                        /* loop through all albums */
                        xmlNodePtr cur = get_first_node_by_name(root,"release");
                        if(cur)
                        {
                            xmlNodePtr cur2 = get_first_node_by_name(cur,"images");
                            if(cur2) {
                                xmlNodePtr cur3 = get_first_node_by_name(cur2,"image");
                                while(cur3 && *path == NULL){
                                    xmlChar *temp = xmlGetProp(cur3, (xmlChar *)"type");

                                    if(temp){
                                        if(xmlStrEqual(temp, (xmlChar *)"primary"))
                                        {
                                            xmlChar *xurl = xmlGetProp(cur3, (xmlChar *)"uri");
                                            xmlChar *size = xmlGetProp(cur3, (xmlChar *)"height");
                                            glong s = (gulong)g_ascii_strtoll((char *)size,NULL, 10);
                                            if(s > psize) {
                                                psize = s;
                                                if(primary) g_free(primary);
                                                primary = g_strdup((char *)xurl);
                                            }
                                            if(xurl) xmlFree(xurl);
                                            if(size) xmlFree(size);
                                        } else if(xmlStrEqual(temp, (xmlChar *)"secondary"))
                                        {
                                            xmlChar *xurl = xmlGetProp(cur3, (xmlChar *)"uri");
                                            xmlChar *size = xmlGetProp(cur3, (xmlChar *)"height");
                                            glong s = (gulong)g_ascii_strtoll((char *)size,NULL, 10);
                                            if(s > ssize) {
                                                ssize = s;
                                                if(secondary) g_free(secondary);
                                                secondary = g_strdup((char *)xurl);
                                            }
                                            if(xurl) xmlFree(xurl);
                                            if(size) xmlFree(size);
                                        }

                                        xmlFree(temp);
                                    }
                                    cur3 = cur3->next;
                                }
                            }
                        }
                    }
                    xmlFreeDoc(doc);
                }
            }
            gmpc_easy_download_clean(&data);
        }
    }
    if(artist_uri) g_free(artist_uri);
    if(primary || secondary)
    {
        gmpc_easy_download((const char *)(primary)?primary:secondary, &data);
        if(data.size){
            FILE *fp = NULL;
            char *imgpath = NULL;
            imgpath = gmpc_get_metadata_filename(META_ALBUM_ART, song, "jpg");
            fp = g_fopen(imgpath, "wb");
            if(fp)
            {
                size_t size;
                size = fwrite(data.data, sizeof(char), data.size,fp);
                if(size != data.size){
                    debug_printf(DEBUG_ERROR, "Something went wrong writing image to file");
                }
                fclose(fp);
            }
            *path = imgpath;
            found = META_DATA_AVAILABLE;
            debug_printf(DEBUG_INFO,"Found cover art, returning\n");
        }
        gmpc_easy_download_clean(&data);
    }
    if(primary) g_free(primary);
    if(secondary) g_free(secondary);
	return found;
}

/** other */
static int discogs_fetch_cover_album_art(mpd_Song *song, char **path)
{
	int retval;
	char *artist = gmpc_easy_download_uri_escape(song->artist);
	char *album = gmpc_easy_download_uri_escape(song->album);
	debug_printf(DEBUG_INFO, "Trying to fetch: %s:%s\n", artist, album);
	retval = __discogs_art_get_album_image(song,artist,album, path);
	g_free(artist);
	g_free(album);
	return retval;
}

static int discogs_fetch_cover_art(mpd_Song *song, char **path)
{
	int retval;
	char *artist = gmpc_easy_download_uri_escape(song->artist);
	retval = __discogs_art_get_artist_image(song,artist,path);
	g_free(artist);
	return retval;
}

static int discogs_fetch_get_image(mpd_Song *song,MetaDataType type, char **path)
{
    int result = 0;
    if(song->artist == NULL || discogs_get_enabled() == FALSE)
    {
        return META_DATA_UNAVAILABLE;
    }
    if(type == META_ARTIST_ART)
    {
        result = discogs_fetch_cover_art(song, path);
        return result;
    }
    else if (type == META_ALBUM_ART && cfg_get_single_value_as_int_with_default(config, "cover-discogs", "fetch-album", TRUE))
    {
        result = discogs_fetch_cover_album_art(song, path);
        return result;
    }

    return META_DATA_UNAVAILABLE;
}

gmpcMetaDataPlugin lf_cover = {
	.get_priority   = discogs_fetch_cover_priority,
    .set_priority   = discogs_fetch_cover_priority_set,
	.get_image      = discogs_fetch_get_image
};

int plugin_api_version = PLUGIN_API_VERSION;

gmpcPlugin plugin = {
	.name           = "DiscoGS Artist and Album Image Fetcher",
	.version        = {PLUGIN_MAJOR_VERSION,PLUGIN_MINOR_VERSION,PLUGIN_MICRO_VERSION},
	.plugin_type    = GMPC_PLUGIN_META_DATA,
	.metadata       = &lf_cover,
	.get_enabled    = discogs_get_enabled,
	.set_enabled    = discogs_set_enabled
};
